require 'balotelli/core/irc'
require 'balotelli/core/router/router'
require 'balotelli/core/router/match'
require 'balotelli/core/router/action'
require 'balotelli/core/priv_msg'
require 'balotelli/core/join'
require 'balotelli/core/kick'
require 'balotelli/core/irc_logger'
require 'balotelli/core/utils'
require 'balotelli/config'
require 'balotelli/thread/pool'

module Balotelli
  class Base
    extend Core::Utils::ClassVariableGetter
    class_variable_get :cache, :cvs, :mutex

    def initialize(block, message, matchdata)
      @message = message
      @block = block
      @matchdata = matchdata
    end

    def execute!
      instance_exec(@message, @matchdata, &@block)
    end

    alias execute execute!

    def out(message_string, destination = nil)
      self.class.privmsg((destination || @message.responder), message_string)
    end

    def kick(params)
      Core::Kick.new(params, self.class).execute!
    end

    def get_user_list(channel)
      execute_in_mutex(channel) do
        self.class.names(channel)
      end
      cache[:user_list][channel]
    end

    def execute_in_mutex(channel)
      cv = ConditionVariable.new
      if cvs.key? channel
        cvs[channel] << cv
      else
        cvs[channel] = [cv]
      end
      yield
      mutex.synchronize { cv.wait(mutex) }
    end
    private :execute_in_mutex

    def method_missing(method, *args, **params, &block)
      if method =~ /__([[[:upper:]]_]+)__/
        self.class.instance_variable_get("@#{Regexp.last_match(1)}".downcase)
      else
        super
      end
    end

    class << self
      include Balotelli::Core::IRC
      include Balotelli::Core::Router

      def setup
        setup_router

        @pool = Thread::Pool.new(Thread::Pool.cpu_count)
        @modules = {}
        @cache = Hash.new { |hash, key| hash[key] = {} }
        @cvs = {}
      end

      def inherited(subclass)
        super

        subclass.setup
      end

      def register(mod, namespace = nil)
        if namespace
          mod.instance_variable_set(:@commands_namespace, namespace.intern)
          mod.instance_exec(namespace) do |n|
            define_method(:commands_namespace) { n.intern }
          end
        end
        @modules[mod.commands_namespace] ||= mod
        @cache[:modules] = Hash[@modules.map { |k, v| [k.to_s, v] }]
      end

      def register_special(mod)
        @routes.merge!(mod.instance_variable_get(:@routes))
      end

      def mod_match(str, priv = false)
        if str =~ /\A[^a-zA-Z0-9\s]?(\S+) ?(.*)/
          if (mod = @cache[:modules][Regexp.last_match(1).downcase])
            mod.match(Regexp.last_match(2), priv)
          end
        elsif str == :join
          match = nil
          @modules.each do |_, m|
            if (match = m.match(:join))
              break(match)
            end
          end
          match
        end
      end

      def run
        while (str = sgets)
          process(str)
        end
      end

      private

      def process(str)
        if str =~ /\APING (.*)\z/
          pong Regexp.last_match(1)
        elsif @on_connect && str =~ %r{End of \/MOTD command}i
          instance_eval(&remove_instance_variable(:@on_connect))
        elsif !(str =~ /(JOIN\s|PRIVMSG\s)/)
          process_table(str)
        else
          @pool.process { process_message(str) }
        end
      end

      def process_message(str)
        if str =~ Core::Utils::PRIV_MSG_REGEX
          process_private_message(str, *Regexp.last_match[1..3])
        elsif str =~ Core::Utils::JOIN_REGEX
          process_join(str, *Regexp.last_match[1..2])
        end
      end

      def process_private_message(str, user, channel, message)
        privacy = !(channel =~ /#.+/)
        if (matched = match(message, privacy)) ||
           (matched = mod_match(message, privacy))
          new_response(user, channel, message, matched, Core::PrivMsg, privacy)
        end
      end

      def process_join(str, user, channel)
        if (matched = match(:join)) || (matched = mod_match(:join))
          new_response(user, channel, nil, matched, Core::Join)
        end
      end


      def new_response(user, channel, message, route_match, klass, priv = nil)
        message = klass.new(*[user, channel, message, priv].compact)
        new(route_match.block, message, route_match.match_data).tap do |response|
          response.extend(route_match.mod_name) if route_match.mod_name.class == ::Module
        end.execute!
      end

      def process_table(string)
        if @cvs.any? && string =~ (regex = Core::Utils.table_regex(@nick))
          channel = Regexp.last_match(2)
          users = Regexp.last_match(3)
          until (str = sgets) =~ %r{End of \/NAME}i
            users << " #{str.match(regex).captures.first}"
          end
          @cache[:user_list][channel] = users.split(' ')
          signal_cvs(channel)
        end
      end

      def signal_cvs(channel)
        @cvs[channel].shift.signal
        @cvs.delete(channel) if @cvs[channel].empty?
      end
    end
  end
end
